#include <cstdio>
#include <cstring> // memset

#include "cpu.hh"

class bisqppu
{
	friend class bisqnes;
	
    BYTE vram[0x2000];
    BYTE *vpage[8];    // 1k pages (0x400)
    BYTE *vrom;
    
    typedef BYTE natab[0x400];
    natab namtab[4];
    BYTE palette[0x20];
    
    BYTE spr_ram[256];
    
    BYTE nametable(unsigned ind, unsigned ofs) const { return namtab[ind][ofs]; }
    BYTE attrtable(unsigned ind, unsigned ofs) const { return namtab[ind][ofs|0x3C0]; }
    
    void writemem(fWORD addr, fBYTE value)
    {
    	addr &= 0x3FFF;
    	if(addr < 0x2000) { if(!vrom)vpage[addr/0x400][addr&0x3FF] = value; return; }
    	if(addr < 0x3F00)
    	{
    		addr &= 0x1FFF;
    		unsigned ind = addr / 0x400;
    		namtab[ind][addr & 0x3FF] = value;
    	}
    	addr &= 0x0F;
    	palette[addr | 0x10] = palette[addr] = value & 0x3F;
    }
    BYTE readmem(fWORD addr) const
    {
    	addr &= 0x3FFF;
    	if(addr < 0x2000) return vpage[addr/0x400][addr&0x3FF];
    	if(addr < 0x3F00)
    	{
    		addr &= 0x1FFF;
    		unsigned ind = addr / 0x400;
    		return namtab[ind][addr & 0x3FF];
    	}
		return palette[addr & 0x1F];
	}
	
public:
	void mapvram()
	{
		for(unsigned a=0; a<8; ++a)
			vpage[a] = &vram[a*0x400];
	}
	void resizevrom(unsigned size)
	{
		if(vrom)delete[] vrom;
		vrom = new BYTE[size];
		for(unsigned a=0; a<8; ++a)
			vpage[a] = &vrom[a*0x400];
	}
	BYTE *vromptr() { return &vrom[0]; }
	
public:
	unsigned dmaptr;
	WORD memptr;
	BYTE reg2000;
	BYTE reg2001;
	BYTE reg2003;
	bool vblank;
	
	unsigned tickpos, scanlid;
	bisqppu() : vrom(NULL)
	{
		dmaptr = 0;
		tickpos = 0;
		scanlid = 0;
		memptr = 0;
		reg2000 = 0x80;
		
		mapvram();
	}
	~bisqppu()
	{
		if(vrom)delete[] vrom;
	}

	bool write(const fWORD addr, fBYTE value)
	{
		switch(addr)
		{
			case 0x2000: reg2000 = value; return true;
			case 0x2001: reg2001 = value; return true;
			case 0x2003: reg2003 = value; return true;
			case 0x2004:
				spr_ram[reg2003++] = value;
				return true;
			case 0x2006:
				memptr = (memptr<<8) | value;
				return true;
			case 0x2007:
				writemem(memptr, value);
				memptr += (reg2000&4) ? 32 : 1;
				return true;
		}
		return false;
	}
	bool read(const fWORD addr, BYTE &value)
	{
		switch(addr)
		{
			case 0x2002:
				if(vblank) { value |= 0x80; vblank = false; }
				return true;
			case 0x2007:
				value = readmem(memptr);
				memptr += (reg2000&4) ? 32 : 1;
				return true;
		}
		return false;
	}
	
	/* We do it *exactly* like NES, no matter how slow it is */
	InterruptType tick()
	{
		if(tickpos==340)
		{
			tickpos = 0;
			++scanlid;
			if(scanlid == 240)
			{
				fprintf(stderr, "!");
				fflush(stderr);
				vblank = true;
			}
			else if(scanlid == 262)
    		{
    			fflush(stderr);
    			scanlid = 0;
    			vblank = false;
    			if(reg2000 & 0x80)
    			{
    				fprintf(stderr, "**NMI**\n");
    				return INT_NMI;
    			}
    		}
    		return INT_NONE;
		}
		if(!(tickpos&1))
		{
			if(tickpos < 256)
			{
				// Draw background 8 pix
			}
			else
			{
				// Fetch sprite for next scanline
			}
		}
		++tickpos;
		return INT_NONE;
	}
};

class bisqpapu
{
public:
	bool write(const fWORD addr, fBYTE value)
	{
		switch(addr)
		{
		}
		return false;
	}
	bool read(const fWORD addr, BYTE &value)
	{
		switch(addr)
		{
		}
		return false;
	}
	
	InterruptType tick()
	{
		
	}
};

class bisqnes
{
private:
	m6502cpu cpu;
	
    bisqppu ppu;
    bisqpapu papu;
	
	BYTE ram[0x800];   // Mirrored four times
	BYTE sram[0x2000]; // Battery RAM
	
	BYTE *rom;
	BYTE rom16count;
	BYTE ctrlbyte;
	BYTE vrom8count;
	BYTE mapper;
	
	BYTE *rompage[8]; // 4k pages (0x1000)
	
public:
	bisqnes() : cpu(1), rom(NULL)
	{
	}
	~bisqnes()
	{
		if(rom)delete[] rom;
	}
	
	void load(const char *fn)
	{
		FILE *fp = fopen(fn, "rb");
		if(!fp)
			return;
			
		if(fgetc(fp)=='N'
		&& fgetc(fp)=='E'
		&& fgetc(fp)=='S'
		&& fgetc(fp)=='\32')
		{
			rom16count = fgetc(fp);
			vrom8count = fgetc(fp);
			ctrlbyte = fgetc(fp);
			mapper = fgetc(fp) | (ctrlbyte>>4);
			fseek(fp, 0x10, SEEK_SET);
			
			if(rom)delete[] rom;
			rom = new BYTE[rom16count * 16384];
			fread(&rom[0], 16384, rom16count, fp);
			
			ppu.resizevrom(vrom8count * 8192);
			fread(ppu.vromptr(), 8192, vrom8count, fp);
		}
		
		fclose(fp);
		
		// ROM pages ok, cpu reset etc.
		poweron();
	}
	
	void run()
	{
		cpu.Run();
	}
	
	void poweron()
	{
		unsigned rom4count = rom16count * 4;
    	for(unsigned a=0; a<4; ++a)
    	{
    		rompage[a]   = &rom[             a  * 0x1000];
    		rompage[a+4] = &rom[(rom4count-4+a) * 0x1000];
    	}
		memset(ram, 0, sizeof ram);
		memset(sram, 0, sizeof sram);
		cpu.Reset();
	}
public:
    inline void Wr(register fWORD Addr, register fBYTE Value)
    {
    	if(Addr < 0x2000) { ram[Addr & 0x7FF] = Value; return; }
    	if(Addr < 0x6000)
    	{
    		if(Addr == 0x4014)
    		{
    			/* DMA transfer source address */
    			Addr = Value * 0x100;
    			/* This _should_ move one byte per 6 tick()s,
    			 * but it's hard to write.
    			 */
    			for(unsigned a=0; a<256; ++a)
    				ppu.spr_ram[a] = Rd(Addr++);
    			return;
    		}
    		if(ppu.write(Addr, Value))return;
    		if(papu.write(Addr, Value))return;
    	}
    	else if(Addr < 0x8000)
    	{
    		if(ctrlbyte&8)
    			sram[Addr & 0x1FFF] = Value;
    	}
    }
    inline BYTE Rd(register fWORD Addr)
    {
    	if(Addr < 0x2000)return ram[Addr & 0x7FF];
    	if(Addr < 0x6000)
    	{
    		BYTE value = 0;
    		if(ppu.read(Addr, value))return value;
    		if(papu.read(Addr, value))return value;
    	}
    	if(Addr < 0x8000)return sram[Addr & 0x1FFF];
    	return rompage[(Addr-0x8000)/0x1000][Addr & 0xFFF];
    }
    inline BYTE Op(register fWORD Addr) const
    {
    	if(Addr < 0x2000)return ram[Addr & 0x7FF];
    	if(Addr < 0x6000)return 0;
    	if(Addr < 0x8000)return sram[Addr & 0x1FFF];
    	Addr -= 0x8000;
    	return rompage[Addr/0x1000][Addr & 0xFFF];
    }
    InterruptType tick()
    {
    	InterruptType tmp = ppu.tick();
    	if(tmp != INT_NONE)return tmp;
    	tmp = papu.tick();
    	if(tmp != INT_NONE)return tmp;
    	
    	return INT_NONE;    	
    }
};

static bisqnes nes;

void Wr6502(register fWORD Addr, register fBYTE Value)
{
	nes.Wr(Addr, Value);
}
BYTE Rd6502(register fWORD Addr)
{
	return nes.Rd(Addr);
}
BYTE Op6502(register fWORD Addr)
{
	return nes.Op(Addr);
}

InterruptType m6502cpu::Loop()
{
	return nes.tick();
}

int main(void)
{
	nes.load("lunar.nes");
	nes.run();
}
